/*
by: Aqrit ( www.bitpatch.com )
date: December 6, 2014
license: Public Domain 

wined3d's ddraw.dll is renamed to wdraw.dll

send every call to wdraw.dll
except opengl32's call to DirectDrawCreate()
which should instead go to system32\ddraw.dll 

The origin of the call is determined by examining the return address on the call stack.
note GetModuleHandleEx( GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS ) requires WinXP
but the code would get ugly(er) without it.

determine which modules the handles correspond to
by examining export virtual addresses

wdraw.lib was created by renaming ddraw.lib
then replacing all occurances of DDRAW.DLL in wdraw.lib  
with WDRAW.DLL
( MSVC don't make it easy... hex-edit the import table post build, if you prefer )
*/
#pragma comment(lib, "wdraw.lib" )

#pragma comment(linker, "/dll")
#pragma comment(linker, "/entry:DllEntryPoint")

/* regular exports */
#pragma comment(linker, "/export:DirectDrawCreate=_my_DirectDrawCreate@12")
/* forwarded exports */
#pragma comment(linker, "/export:DirectDrawCreateClipper=wdraw.DirectDrawCreateClipper")
#pragma comment(linker, "/export:DirectDrawCreateEx=wdraw.DirectDrawCreateEx")
#pragma comment(linker, "/export:DirectDrawEnumerateA=wdraw.DirectDrawEnumerateA")
#pragma comment(linker, "/export:DirectDrawEnumerateExA=wdraw.DirectDrawEnumerateExA")
#pragma comment(linker, "/export:DirectDrawEnumerateExW=wdraw.DirectDrawEnumerateExW")
#pragma comment(linker, "/export:DirectDrawEnumerateW=wdraw.DirectDrawEnumerateW")
#pragma comment(linker, "/export:DllCanUnloadNow=wdraw.DllCanUnloadNow")
#pragma comment(linker, "/export:DllGetClassObject=wdraw.DllGetClassObject")
#pragma comment(linker, "/export:DllRegisterServer=wdraw.DllRegisterServer")
#pragma comment(linker, "/export:DllUnregisterServer=wdraw.DllUnregisterServer")

#include <windows.h>
#include <ddraw.h>

typedef HRESULT (__stdcall* DirectDrawCreate_t)( GUID*, LPDIRECTDRAW*, IUnknown* );
extern "C" void* _ReturnAddress(void);
#pragma intrinsic( _ReturnAddress )

/* must get this thru GetProcAddress, can't trust all linkers */
DirectDrawCreate_t g_my_DDCreate;

BOOL __stdcall DllEntryPoint( HINSTANCE hDll, DWORD dwReason, LPVOID lpvReserved ){
	if( dwReason == DLL_PROCESS_ATTACH )
	{
		g_my_DDCreate = (DirectDrawCreate_t) GetProcAddress( hDll, "DirectDrawCreate" );
		DisableThreadLibraryCalls( hDll );
	}
	return TRUE;
}

#pragma intrinsic( strcat )
extern "C" HRESULT __stdcall my_DirectDrawCreate( GUID* lpGUID, LPDIRECTDRAW* lplpDD, IUnknown* pUnkOuter )
{
	HMODULE	hCaller;
	HMODULE hSysDDraw;
	DirectDrawCreate_t pDirectDrawCreate;
	int n;
	char szPath[ MAX_PATH ];

	pDirectDrawCreate = NULL;

	GetModuleHandleEx( 
		GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
		(LPCSTR) _ReturnAddress(),
		&hCaller 
	);

	/* if the caller exports wglGetProcAddress then it is probably opengl32.dll */
	if( hCaller && GetProcAddress( hCaller, "wglGetProcAddress" ) )
	{
		n = GetSystemDirectory( szPath, MAX_PATH - 11 );
		if( n && ( n <= MAX_PATH - 11 ) )
		{
			strcat( szPath, "\\ddraw.dll" );

			/* not worried about ref count, not planning to unload this anyways */
			hSysDDraw = LoadLibrary( szPath );
			if( hSysDDraw ) pDirectDrawCreate = (DirectDrawCreate_t) GetProcAddress( hSysDDraw, "DirectDrawCreate");
			if( pDirectDrawCreate == g_my_DDCreate ) pDirectDrawCreate = NULL;
		}
	}

	if( pDirectDrawCreate == NULL )
		pDirectDrawCreate = &DirectDrawCreate;

	return pDirectDrawCreate( (GUID*)lpGUID, lplpDD, pUnkOuter );
}


